from django.db import models
from enum import Enum


class VERBMAP(Enum):
    POST = "POST"
    GET = "GET"

    @classmethod
    def choices(cls):
        return tuple((i.name, i.value) for i in cls)


class SOAPRequestLogger(models.Model):
    """
    Logs SOAP requests
    """
    date_sent = models.DateTimeField(auto_now_add=True)
    method = models.CharField(
        max_length=4,
        choices=VERBMAP.choices(),
        default=VERBMAP.POST
    )
    url = models.URLField(max_length=255, null=False)
    headers = models.TextField(blank=True, null=False)
    body = models.TextField(blank=True, null=True)
    request_time_ms = models.IntegerField(blank=True, null=True)

    def __str__(self) -> str:
        return f"{self.date_sent} - {self.url} - {self.request_time_ms}"

    class Meta:
        ordering = ('-date_sent',)


class SOAPResponseLogger(models.Model):
    """
    Logs SOAP responses
    """
    date_received = models.DateTimeField(blank=True, null=True)
    request_id = models.ForeignKey(SOAPRequestLogger, on_delete=models.CASCADE)
    status = models.IntegerField(blank=True, null=True)
    headers = models.TextField(blank=True, null=True)
    body = models.TextField(blank=True, null=True)

    def __str__(self) -> str:
        return f"{self.date_received} - {self.request_id} - {self.status}"

    class Meta:
        ordering = ('-date_received',)


# Generated by Django 3.2.6 on 2021-08-07 05:23

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='SOAPRequestLogger',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_sent', models.DateTimeField(auto_now_add=True)),
                ('method', models.CharField(max_length=10)),
                ('url', models.URLField(max_length=255)),
                ('headers', models.TextField(blank=True, null=True)),
                ('body', models.TextField(blank=True, null=True)),
                ('request_time_ms', models.IntegerField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='SOAPResponseLogger',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_received', models.DateTimeField(blank=True, null=True)),
                ('status', models.IntegerField(blank=True, null=True)),
                ('headers', models.TextField(blank=True, null=True)),
                ('body', models.TextField(blank=True, null=True)),
                ('request_id', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_soap.soaprequestlogger')),
            ],
        ),
    ]



from django.contrib import admin
from django_soap.models import *

# Register your models here.
class SOAPRequstLoggerAdmin(admin.ModelAdmin):
    ...

class SOAPResponseLoggerAdmin(admin.ModelAdmin):
    ...

admin.site.register(SOAPResponseLogger, SOAPResponseLoggerAdmin)
admin.site.register(SOAPRequestLogger, SOAPRequstLoggerAdmin)


from django_soap.utils.SOAPHandlerBase import SOAPHandlerBase


class SOAPClient(SOAPHandlerBase):
    """
    Sends request to server and returns response
    """
    def __init__(self, *args, **kwargs):
        super(SOAPClient, self).__init__(*args, **kwargs)
    
    def send_request(self, envelope_path: str, envelope_attributes):
        return self.post(envelope_path, envelope_attributes)
        


import requests
import json
import xmltodict
from django_soap.models import (
    SOAPRequestLogger,
    SOAPResponseLogger,
    VERBMAP
)
from collections import OrderedDict
from django.template.loader import get_template, render_to_string
from datetime import datetime, timezone
from django_soap.utils import exceptions
from django.db.models.query import QuerySet


class SoapResult:
    """
    Parses xml response
    """
    def __init__(self, data, status_code):
        if isinstance(data, str):
            self._dict = dict(xmltodict.parse(data))
        elif isinstance(data, dict):
            self._dict = data
        elif isinstance(data, list):
            self._dict = OrderedDict({'Response': data})
        
        self.xml = data
        self.status_code = status_code
        self.value = self._dict

    def _parse_path(self, path: str) -> list:
        return path.split('/')

    def xml_to_json(self):
        return json.dumps(self._dict)
    
    # def has(self, path: str):
    #     try:
    #         self._get(self._dict, self._parse_path(path))
    #     except exceptions.NodeNotFound:
    #         return False
    #     return True
    
    def has(self, value):
        if value in list(self._dict.keys()):
            return True
        return False
    
    def get(self, path, default_value=None):
        try:
            result = self._get(self._dict, self._parse_path(path))
        except exceptions.NodeNotFoundError:
            return default_value
        if isinstance(result, (dict, OrderedDict)):
            return SoapResult(result, self.status_code)
        return result

    def get_list(self, path):
        try:
            result = self._get(self._dict, self._parse_path(path))
        except exceptions.NodeNotFoundError:
            return SoapResult([], self.status_code)
        if isinstance(result, list):
            return SoapResult(result, self.status_code)
        if isinstance(result, (dict, OrderedDict)):
            return SoapResult([result], self.status_code)
        return

    def _find(self, d, key):
        for k, v in d.items():
            if isinstance (v, OrderedDict):
                for found in self._find(v, key):
                    yield found
            if k == key:
                yield v

    def find(self, key):
        return SoapResult(next(self._find(self._dict, key)), self.status_code)

    def get_value(self, path, default_value=False):
        result = self.get(path, default_value)
        if isinstance(result, SoapResult):
            raise exceptions.NotTextFound
        return result

    def _get(self, data, keys):
        try:
            return self._get(data[keys[0]], keys[1:]) if keys else data
        except KeyError:
            raise exceptions.NodeNotFoundError


class SOAPHandlerBase():
    """
    This class is the base class for all the SOAP handlers.
    """
    url = None
    headers = {}

    def __init__(self, *args, **kwargs):
        self.__dict__.update(kwargs)
    
    def _send_request(self, verb: VERBMAP, url: str, data, headers):
        """
        Sends the request to the server.
        """

        if url is None and self.url is None:
            raise exceptions.URLNotFound
        
        if headers is None and (self.headers is {} or self.headers is None):
            raise exceptions.HeaderNotFound
        
        req_log = SOAPRequestLogger.objects.create(
            method=verb.value,
            url=url or self.url,
            headers=str(headers or self.headers),
            body=data,            
        )
        
        r = requests.post(url, data=data, headers=headers)
        response = self._receive_response(req_log, r, log)
        
        req_log.request_time_ms = (
            datetime.now(timezone.utc) - req_log.date_sent
        ).microseconds
        req_log.save()

        return response
    
    def _handle_fault_codes(self, response):
        """
        Handles the fault codes.
        """
        if response.status_code != 200:
            error = (
                response
                    .find('soap:Fault')
                    .get_value('faultstring')
            )
            raise Exception(
                f"{error.split(':')[1]} \n"
            )
        return response.status_code

    def _handle_status_codes(self, response):
        """
        Reads status code and returns value, but if it's a 
        bad resoponse, raises an exception.
        """
        if response.status_code < 400:
            return response.status_code

        # SOAP_STATUS_EXCEPTIONS = {
        #     400: 'Bad Request',
        #     401: 'Unauthorized',
        #     403: 'Forbidden',
        #     404: 'Not Found',
        #     405: 'Method Not Allowed',
        #     406: 'Not Acceptable',
        #     500: 'Internal Server Error',
        #     501: 'Not Implemented',
        #     502: 'Bad Gateway',
        #     503: 'Service Unavailable',
        #     504: 'Gateway Timeout'
        # }

        # try:
        #     ...
        # except:
        #     raise exceptions.SOAPStatusException(SOAP_STATUS_EXCEPTIONS[response.status_code])
        
        return response.status_code
    
    def _receive_response(self, request_logger, response) -> SoapResult:
        """
        Updates the SOAPResponseLogger with response info and
        parses the response to a SOAP Result.
        """
        status_code = self._handle_status_codes(response)
        
        SOAPResponseLogger.objects.create(
            request_id=request_logger,
            status=status_code,
            headers=str(response.headers),
            body=str(response.content),
        )

        result = SoapResult(response.text, response.status_code)

        # if status_code != 200:
        #     raise exceptions.SOAPStatusException(
        #         message=result.get(''
        #     )

        return result

    def _prepare_envelope(self, envelope_path: str, envelope_attributes: dict):
        """
        This method prepares the envelope / SOAP request.
        """
        return render_to_string(
            self._get_envelope(envelope_path),
            context=envelope_attributes
        )
    
    def _extend_attributes(self, attributes: dict) -> dict:
        """
        Grabs attributes from class and newely inputed attiibutes.
        The newly inputed attributes overwrite the class attributes.
        """
        
        return (lambda f=self.__dict__.copy(): (f.update(attributes), f)[1])()
    
    def _get_envelope(self, path: str) -> str:
        """
        Looks for the XML file in the templates directory.
        """
        try:
            template = get_template(path)
        except exceptions.TemplateNotFound:
            raise exceptions.TemplateNotFound
        
        return path

    def post(self, envelope_path=None, envelope_attributes=None, log=False) -> SoapResult:
        """
        Posts the request to the server.
        """
        # convert all querysets to lists
        for k, v in envelope_attributes.items():
            if isinstance(v, QuerySet):
                envelope_attributes[k] = list(v.values())

        if isinstance(envelope_attributes, QuerySet):
            envelope_attributes = list(envelope_attributes.values())
    
        payload = self._prepare_envelope(
            envelope_path, 
            self._extend_attributes(envelope_attributes or {})
        )

        result = self._send_request(
            VERBMAP.POST,
            envelope_attributes.get('url') or self.url,
            data=payload,
            headers=self.headers,
        )
        return result
    
    def get(self, envelope_path=None, envelope_attributes=None) -> SoapResult:
        # NOTE I don't think SOAP does GET requests??
        ...



class NodeNotFound(Exception):
    """
    Exception raised when a node is not found in the graph.
    """
    ...

class URLNotFound(Exception):
    """
    Exception raised when a URL is not found for request.
    """
    ...

class HeaderNotFound(Exception):
    """
    Exception raised when a header is not found for request.
    """
    ...

class SOAPStatusException(Exception):
    """
    Exception raised when a SOAP request returns an error.
    """
    ...

class TemplateNotFound(Exception):
    """
    Exception raised when a template is not found.
    """
    ...

class FaultCodeException(Exception):
    """
    Exception raised when a SOAP request returns a fault code.
    """
    def __init__(self, message="", code=None, detail=None):
        super(Exception, self).__init__(message)
        self.message = message
        self.code = code
        self.detail = detail

class NodeNotFoundError(Exception):
    """
    Exception raised when XML element not found
    """
    ...


# Generated by Django 3.2.6 on 2021-08-07 06:21

from django.db import migrations, models
import django_soap.models


class Migration(migrations.Migration):

    dependencies = [
        ('django_soap', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='soaprequestlogger',
            options={'ordering': ('-date_sent',)},
        ),
        migrations.AlterModelOptions(
            name='soapresponselogger',
            options={'ordering': ('-date_received',)},
        ),
        migrations.AlterField(
            model_name='soaprequestlogger',
            name='method',
            field=models.CharField(choices=[('POST', 'POST'), ('GET', 'GET')], default=django_soap.models.VERBMAP['POST'], max_length=4),
        ),
    ]
